#!/usr/bin/env nextflow

nextflow.enable.dsl = 2

// import relevant modules
include { FASTQC } from '../modules/fastqc'
include { FASTQC_TRIM } from '../modules/fastqc_trim'
include { TRIMMOMATIC } from '../modules/trimmomatic'
include { FILTER_READS } from '../modules/filter_reads'
include { BOWTIE2_ALIGN } from '../modules/bowtie2_align'
include { SAMTOOLS_FILTERUNMAPPED } from '../modules/samtools_filterunmapped'
include { PICARD_SORTSAM } from '../modules/picard_sortsam'
include { PICARD_DSSORT } from '../modules/picard_dssort'
include { PICARD_MARKDUPLICATES } from '../modules/picard_markduplicates'
include { SAMTOOLS_DOWNSAMPLE } from '../modules/samtools_downsample'
include { PICARD_INSERTMETRICS } from '../modules/picard_insertmetrics'
include { COLLECT_BAMS } from '../modules/collect_bams'
include { COVERAGE } from '../modules/coverage'
include { CLEAN_COUNTS } from '../modules/clean_counts'
include { CORRELATION } from '../modules/correlation'
include { MULTIQC_REPORT } from '../modules/multiqc_report'
include { DNA_QC_REPORT } from '../modules/dna_qc_report'
include { SAMTOOLS_BAMTOBED } from '../modules/samtools_bamtobed'
include { GINKGO_BINUNSORTED } from '../modules/ginkgo_binunsorted'
include { CONCAT_BINNED_READS } from '../modules/concat_binned_reads'
include { GINKGO_PROCESS } from '../modules/ginkgo_process'
include { DNA_REPORT } from '../modules/dna_report'

workflow DNA_ANALYZE_WORKFLOW {
    
    take:
        input_fastq_ch
        demux_counts_ch
    
    main:

        // check if bowtie2 index is setup correctly

        def bowtie2_index = file("${params.genome_dir}/${params.genome}/bowtie2_index")
        def ginkgo_ref_index = file("${params.genome_dir}/${params.genome}/ginkgo_ref")
        def genome_fasta = file("${params.genome_dir}/${params.genome}/${params.genome}.genome.fa")

        // only validate genome paths if it is not in a s3 bucket
        if ( !params.genome_dir.startsWith('s3:') ) {
            if ( !bowtie2_index.exists() ) {
                exit 1, "Bowtie2 Index path does not exist, please run cogent dna add_genome first"
            } 
            if ( !ginkgo_ref_index.exists() ) {
                exit 1, "Ginkgo Index path does not exist, please run cogent dna add_genome first"
            } 
        }

        Channel.fromPath(bowtie2_index)
            .set{ bowtie2_index_ch }
        
        Channel.fromPath(ginkgo_ref_index)
            .set{ ginkgo_index_ch }

        Channel.fromPath(genome_fasta)
                .set{ genome_fasta_ch }
            
        Channel.fromPath(params.barcodes_file)
            .set{ bcs_file_ch }

        FASTQC(input_fastq_ch)

        raw_fastqc_multiqc_ch = Channel.empty()
        raw_fastqc_multiqc_ch = FASTQC.out.fastqc_zip

        input_fastq_ch
            .map { it[1] } 
            .collect()
            .set { reads_ch }


        FILTER_READS(reads_ch, demux_counts_ch)

        def filtered_reads = FILTER_READS.out.filtered_reads

        filtered_reads
            .collect()
            .flatten()
            .map{ file ->
                    def sample = file.name.replaceAll("_R[12].fastq.gz", "")
                    tuple(sample, file)
            }
            .groupTuple()
            .set{ trimmomatic_input_ch }


        TRIMMOMATIC(trimmomatic_input_ch)

        trimmomatic_multiqc_ch = Channel.empty()
        trimmomatic_multiqc_ch = TRIMMOMATIC.out.log

        def fastq_trimmed = TRIMMOMATIC.out.trimmed_reads

        fastq_trimmed
            .collect()
            .flatten()
            .map{ file ->
                    def sample = file.name.replaceAll("_R[12].fastq.gz", "")
                    tuple(sample, file)
            }
            .groupTuple()
            .set{ fastq_trimmed_ch }
        
        FASTQC_TRIM(fastq_trimmed_ch)

        trim_fastqc_multiqc_ch = Channel.empty()
        trim_fastqc_multiqc_ch = FASTQC_TRIM.out.fastqc_zip

        fastq_trimmed_ch
            .combine(bowtie2_index_ch)
            .set{ bowtie2_align_input_ch }

        BOWTIE2_ALIGN(bowtie2_align_input_ch)

        bowtie2_multiqc_ch = Channel.empty()
        bowtie2_multiqc_ch = BOWTIE2_ALIGN.out.log_file

        PICARD_SORTSAM(BOWTIE2_ALIGN.out.bam)
        sortsam_multiqc_ch = Channel.empty()
        sortsam_multiqc_ch = PICARD_SORTSAM.out.log_file
        
        PICARD_SORTSAM.out.sorted_bam
            .collect(flat: false)
            .flatMap()
            .set{ picard_markduplicates_input_ch }

        PICARD_MARKDUPLICATES(picard_markduplicates_input_ch)
        
        markdup_metrics_multiqc_ch = Channel.empty()
        markdup_metrics_multiqc_ch = PICARD_MARKDUPLICATES.out.mkdup_metrics

        markdup_multiqc_ch = Channel.empty()
        markdup_multiqc_ch = PICARD_MARKDUPLICATES.out.mkdup_log

        SAMTOOLS_FILTERUNMAPPED(BOWTIE2_ALIGN.out.bam)

        SAMTOOLS_DOWNSAMPLE(SAMTOOLS_FILTERUNMAPPED.out.mapped_sam)

        SAMTOOLS_DOWNSAMPLE.out.mapped_sam
            .collect(flat: false)
            .flatMap()
            .set{ picard_dssort_input_ch }

        PICARD_DSSORT(picard_dssort_input_ch)
        
        dssort_multiqc_ch = Channel.empty()
        dssort_multiqc_ch = PICARD_DSSORT.out.log_file

        PICARD_DSSORT.out.sorted_bam
            .collect(flat: false)
            .flatMap()
            .set{ picard_insertmetrics_input_ch }

        PICARD_INSERTMETRICS(picard_insertmetrics_input_ch)

        insertmetrics_multiqc_ch = Channel.empty()
        insertmetrics_multiqc_ch = PICARD_INSERTMETRICS.out.insert_metrics

        correlation_heatmap_multiqc_ch = Channel.empty()
        correlation_matrix_multiqc_ch = Channel.empty()

        COLLECT_BAMS(PICARD_DSSORT.out.sorted_bam, PICARD_DSSORT.out.sorted_bam_index)

        COLLECT_BAMS.out.staged_bam
            .collect()
            .set{ coverage_input_bam_ch }

        COLLECT_BAMS.out.staged_bam_index
            .collect()
            .set { coverage_input_bam_index_ch }
        
        COVERAGE(
            coverage_input_bam_ch, 
            coverage_input_bam_index_ch)

        CLEAN_COUNTS(COVERAGE.out.counts_table)

        correlation_met_ch = Channel.of('pearson')

        CORRELATION(COVERAGE.out.read_counts, correlation_met_ch)

        correlation_heatmap_multiqc_ch = CORRELATION.out.correlation_heatmap_no
        correlation_matrix_multiqc_ch = CORRELATION.out.matrix_no    

        MULTIQC_REPORT(
            raw_fastqc_multiqc_ch.collect{it[1]}.ifEmpty([]),
            trimmomatic_multiqc_ch.collect{it[1]}.ifEmpty([]),
            trim_fastqc_multiqc_ch.collect{it[1]}.ifEmpty([]),
            bowtie2_multiqc_ch.collect{it[1]}.ifEmpty([]),
            sortsam_multiqc_ch.collect{it[1]}.ifEmpty([]),
            dssort_multiqc_ch.collect{it[1]}.ifEmpty([]),
            markdup_multiqc_ch.collect{it[1]}.ifEmpty([]),
            markdup_metrics_multiqc_ch.collect{it[1]}.ifEmpty([]),
            insertmetrics_multiqc_ch.collect{{it[1]}}.ifEmpty([]),
            correlation_heatmap_multiqc_ch.collect{it[1]}.ifEmpty([]),
            correlation_matrix_multiqc_ch.collect{it[1]}.ifEmpty([])
        )

        DNA_QC_REPORT(
            MULTIQC_REPORT.out.data,
            demux_counts_ch, 
            bcs_file_ch,
            CLEAN_COUNTS.out.clean_counts,
            correlation_matrix_multiqc_ch
        )

        SAMTOOLS_BAMTOBED(PICARD_MARKDUPLICATES.out.mkdup_bam)

        SAMTOOLS_BAMTOBED.out.bed
            .combine(ginkgo_index_ch)
            .set{ ginkgo_binunsorted_input_ch }

        def read_length_actual = ''

        if ("${params.read_length}" == "76bp"){
            read_length_actual = '76'
        } else if ("${params.read_length}" == "151bp"){
            read_length_actual = '151'
        } else{
            exit 1, "Unsupported read length, please choose one of 76bp or 151bp"
        }
        
        def bin_size_actual = ''

        if ("${params.bin_size}" == "500kb"){
            bin_size_actual = '500000'
        } else if ("${params.bin_size}" == "1mb"){
            bin_size_actual = '1000000'
        } else{
            exit 1, "Unsupported bin size, please choose one of 500kb or 1Mb"
        }

        GINKGO_BINUNSORTED(ginkgo_binunsorted_input_ch, bin_size_actual, read_length_actual)

        GINKGO_BINUNSORTED.out.mapped
            .collect()
            .set{ concat_binned_reads_ch }

        CONCAT_BINNED_READS(concat_binned_reads_ch)

        GINKGO_PROCESS(CONCAT_BINNED_READS.out.data, ginkgo_index_ch, bin_size_actual, read_length_actual)

        DNA_REPORT(
            ginkgo_index_ch,
            CONCAT_BINNED_READS.out.data,
            GINKGO_PROCESS.out.seg_copy,
            bin_size_actual,
            read_length_actual
        )

}
